@@ -2,36 +2,14 @@
 #include <fcntl.h>
 #include <string.h>
 #include <assert.h>
+#include <dirent.h>
 #include <sys/stat.h>
 #include <sys/mman.h>
+#include <sys/types.h>
 
 #include "fio.h"
 #include "os.h"
 
-int open_file(struct thread_data *td, struct fio_file *f, int flags, int perm)
-{
-	if (flags & O_CREAT)
-		f->fd = open(f->file_name, flags, perm);
-	else
-		f->fd = open(f->file_name, flags);
-
-	if (f->fd != -1) {
-		td->nr_open_files++;
-		return 0;
-	}
-
-	return 1;
-}
-
-void close_file(struct thread_data *td, struct fio_file *f)
-{
-	if (f->fd != -1) {
-		close(f->fd);
-		f->fd = -1;
-		td->nr_open_files--;
-	}
-}
-
 /*
  * Check if the file exists and it's large enough.
  */
@@ -39,7 +17,8 @@ static int file_ok(struct thread_data *td, struct fio_file *f)
 {
 	struct stat st;
 
-	if (td->filetype != FIO_TYPE_FILE)
+	if (f->filetype != FIO_TYPE_FILE ||
+	    (td->io_ops->flags & FIO_DISKLESSIO))
 		return 0;
 
 	if (lstat(f->file_name, &st) == -1)
@@ -121,29 +100,56 @@ err:
 static int create_files(struct thread_data *td)
 {
 	struct fio_file *f;
-	int i, err, need_create;
+	int err, need_create, can_extend;
+	unsigned long long total_file_size;
+	unsigned int i, new_files;
 
-	for_each_file(td, f, i)
-		f->file_size = td->total_file_size / td->nr_files;
+	new_files = 0;
+	total_file_size = td->total_file_size;
+	for_each_file(td, f, i) {
+		unsigned long long s;
+
+		f->file_offset = td->start_offset;
+
+		if (f->filetype != FIO_TYPE_FILE)
+			continue;
+
+		if (f->flags & FIO_FILE_EXISTS) {
+			s = f->file_size;
+			if (s > total_file_size)
+				s = total_file_size;
+
+			total_file_size -= s;
+		} else
+			new_files++;
+	}
 
 	/*
 	 * unless specifically asked for overwrite, let normal io extend it
 	 */
-	if (!td->overwrite)
+	can_extend = !td->overwrite && !(td->io_ops->flags & FIO_NOEXTEND);
+	if (can_extend)
 		return 0;
 
 	need_create = 0;
-	if (td->filetype == FIO_TYPE_FILE) {
-		for_each_file(td, f, i) {
-			int file_there = !file_ok(td, f);
+	for_each_file(td, f, i) {
+		int file_there;
 
-			if (file_there && td_write(td) && !td->overwrite) {
-				unlink(f->file_name);
-				file_there = 0;
-			}
+		if (f->filetype != FIO_TYPE_FILE)
+			continue;
+		if (f->flags & FIO_FILE_EXISTS)
+			continue;
 
-			need_create += !file_there;
+		f->file_size = total_file_size / new_files;
+
+		file_there = !file_ok(td, f);
+
+		if (file_there && td_write(td) && !td->overwrite) {
+			unlink(f->file_name);
+			file_there = 0;
 		}
+
+		need_create += !file_there;
 	}
 
 	if (!need_create)
@@ -157,18 +163,20 @@ static int create_files(struct thread_data *td)
 
 	temp_stall_ts = 1;
 	fprintf(f_out, "%s: Laying out IO file(s) (%u x %LuMiB == %LuMiB)\n",
-				td->name, td->nr_uniq_files,
-				(td->total_file_size >> 20) / td->nr_uniq_files,
-				td->total_file_size >> 20);
+				td->name, new_files,
+				(total_file_size >> 20) / new_files,
+				total_file_size >> 20);
 
 	err = 0;
 	for_each_file(td, f, i) {
 		/*
 		 * Only unlink files that we created.
 		 */
-		f->unlink = 0;
+		f->flags &= ~FIO_FILE_UNLINK;
 		if (file_ok(td, f)) {
-			f->unlink = td->unlink;
+			if (td->unlink)
+				f->flags |= FIO_FILE_UNLINK;
+
 			err = create_file(td, f);
 			if (err)
 				break;
@@ -226,9 +234,10 @@ static int get_file_size(struct thread_data *td, struct fio_file *f)
 {
 	int ret = 0;
 
-	if (td->filetype == FIO_TYPE_FILE)
-		ret = file_size(td, f);
-	else if (td->filetype == FIO_TYPE_BD)
+	if (f->filetype == FIO_TYPE_FILE) {
+		if (!(f->flags & FIO_FILE_EXISTS))
+			ret = file_size(td, f);
+	} else if (f->filetype == FIO_TYPE_BD)
 		ret = bdev_size(td, f);
 	else
 		f->real_file_size = -1;
@@ -248,18 +257,19 @@ int file_invalidate_cache(struct thread_data *td, struct fio_file *f)
 {
 	int ret = 0;
 
+	if (td->odirect)
+		return 0;
+
 	/*
 	 * FIXME: add blockdev flushing too
 	 */
-	if (td->io_ops->flags & FIO_MMAPIO)
+	if (f->mmap)
 		ret = madvise(f->mmap, f->file_size, MADV_DONTNEED);
-	else if (td->filetype == FIO_TYPE_FILE) {
-		if (!td->odirect)
-			ret = fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_DONTNEED);
-	} else if (td->filetype == FIO_TYPE_BD) {
-		if (!td->odirect)
-			ret = blockdev_invalidate_cache(f->fd);
-	} else if (td->filetype == FIO_TYPE_CHAR)
+	else if (f->filetype == FIO_TYPE_FILE) {
+		ret = fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_DONTNEED);
+	} else if (f->filetype == FIO_TYPE_BD) {
+		ret = blockdev_invalidate_cache(f->fd);
+	} else if (f->filetype == FIO_TYPE_CHAR)
 		ret = 0;
 
 	if (ret < 0) {
@@ -270,100 +280,16 @@ int file_invalidate_cache(struct thread_data *td, struct fio_file *f)
 	return ret;
 }
 
-static int __setup_file_mmap(struct thread_data *td, struct fio_file *f)
+void generic_close_file(struct thread_data fio_unused *td, struct fio_file *f)
 {
-	int flags;
-
-	if (td_rw(td))
-		flags = PROT_READ | PROT_WRITE;
-	else if (td_write(td)) {
-		flags = PROT_WRITE;
-
-		if (td->verify != VERIFY_NONE)
-			flags |= PROT_READ;
-	} else
-		flags = PROT_READ;
-
-	f->mmap = mmap(NULL, f->file_size, flags, MAP_SHARED, f->fd, f->file_offset);
-	if (f->mmap == MAP_FAILED) {
-		f->mmap = NULL;
-		td_verror(td, errno, "mmap");
-		return 1;
-	}
-
-	if (td->invalidate_cache && file_invalidate_cache(td, f))
-		return 1;
-
-	if (!td_random(td)) {
-		if (madvise(f->mmap, f->file_size, MADV_SEQUENTIAL) < 0) {
-			td_verror(td, errno, "madvise");
-			return 1;
-		}
-	} else {
-		if (madvise(f->mmap, f->file_size, MADV_RANDOM) < 0) {
-			td_verror(td, errno, "madvise");
-			return 1;
-		}
-	}
-
-	return 0;
+	close(f->fd);
+	f->fd = -1;
 }
 
-static int setup_files_mmap(struct thread_data *td)
-{
-	struct fio_file *f;
-	int i, err = 0;
-
-	for_each_file(td, f, i) {
-		err = __setup_file_mmap(td, f);
-		if (err)
-			break;
-	}
-
-	return err;
-}
-
-static int __setup_file_plain(struct thread_data *td, struct fio_file *f)
-{
-	if (td->invalidate_cache && file_invalidate_cache(td, f))
-		return 1;
-
-	if (!td_random(td)) {
-		if (fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_SEQUENTIAL) < 0) {
-			td_verror(td, errno, "fadvise");
-			return 1;
-		}
-	} else {
-		if (fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_RANDOM) < 0) {
-			td_verror(td, errno, "fadvise");
-			return 1;
-		}
-	}
-
-	return 0;
-}
-
-static int setup_files_plain(struct thread_data *td)
-{
-	struct fio_file *f;
-	int i, err = 0;
-
-	for_each_file(td, f, i) {
-		err = __setup_file_plain(td, f);
-		if (err)
-			break;
-	}
-
-	return err;
-}
-
-static int setup_file(struct thread_data *td, struct fio_file *f)
+int generic_open_file(struct thread_data *td, struct fio_file *f)
 {
 	int flags = 0;
 
-	if (td->io_ops->flags & FIO_SELFOPEN)
-		return 0;
-
 	if (td->odirect)
 		flags |= OS_O_DIRECT;
 	if (td->sync_io)
@@ -372,21 +298,17 @@ static int setup_file(struct thread_data *td, struct fio_file *f)
 	if (td_write(td) || td_rw(td)) {
 		flags |= O_RDWR;
 
-		if (td->filetype == FIO_TYPE_FILE) {
-			if (!td->overwrite)
-				flags |= O_TRUNC;
-
+		if (f->filetype == FIO_TYPE_FILE)
 			flags |= O_CREAT;
-		}
 
-		open_file(td, f, flags, 0600);
+		f->fd = open(f->file_name, flags, 0600);
 	} else {
-		if (td->filetype == FIO_TYPE_CHAR)
+		if (f->filetype == FIO_TYPE_CHAR)
 			flags |= O_RDWR;
 		else
 			flags |= O_RDONLY;
 
-		open_file(td, f, flags, 0);
+		f->fd = open(f->file_name, flags);
 	}
 
 	if (f->fd == -1) {
@@ -395,33 +317,55 @@ static int setup_file(struct thread_data *td, struct fio_file *f)
 		td_verror(td, __e, "open");
 		if (__e == EINVAL && td->odirect)
 			log_err("fio: destination does not support O_DIRECT\n");
+		if (__e == EMFILE)
+			log_err("fio: try reducing/setting openfiles (failed at %u of %u)\n", td->nr_open_files, td->nr_files);
 		return 1;
 	}
 
-	if (get_file_size(td, f)) {
-		close_file(td, f);
-		return 1;
+	if (get_file_size(td, f))
+		goto err;
+
+	if (td->invalidate_cache && file_invalidate_cache(td, f))
+		goto err;
+
+	if (!td_random(td)) {
+		if (fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_SEQUENTIAL) < 0) {
+			td_verror(td, errno, "fadvise");
+			goto err;
+		}
+	} else {
+		if (fadvise(f->fd, f->file_offset, f->file_size, POSIX_FADV_RANDOM) < 0) {
+			td_verror(td, errno, "fadvise");
+			goto err;
+		}
 	}
 
 	return 0;
+err:
+	close(f->fd);
+	return 1;
 }
 
 int open_files(struct thread_data *td)
 {
 	struct fio_file *f;
-	int i, err = 0;
+	unsigned int i;
+	int err = 0;
 
 	for_each_file(td, f, i) {
-		err = setup_file(td, f);
+		err = td_io_open_file(td, f);
 		if (err)
 			break;
+
+		if (td->open_files == td->nr_open_files)
+			break;
 	}
 
 	if (!err)
 		return 0;
 
 	for_each_file(td, f, i)
-		close_file(td, f);
+		td_io_close_file(td, f);
 
 	return err;
 }
@@ -429,7 +373,8 @@ int open_files(struct thread_data *td)
 int setup_files(struct thread_data *td)
 {
 	struct fio_file *f;
-	int err, i;
+	unsigned int i;
+	int err;
 
 	/*
 	 * if ioengine defines a setup() method, it's responsible for
@@ -464,13 +409,8 @@ int setup_files(struct thread_data *td)
 
 	td->total_io_size = td->io_size * td->loops;
 
-	if (td->io_ops->flags & FIO_MMAPIO)
-		err = setup_files_mmap(td);
-	else
-		err = setup_files_plain(td);
-
 	for_each_file(td, f, i)
-		close_file(td, f);
+		td_io_close_file(td, f);
 
 	return err;
 }
@@ -478,26 +418,135 @@ int setup_files(struct thread_data *td)
 void close_files(struct thread_data *td)
 {
 	struct fio_file *f;
-	int i;
+	unsigned int i;
 
 	for_each_file(td, f, i) {
-		if (!td->filename && f->unlink &&
-		    td->filetype == FIO_TYPE_FILE) {
+		if (!f->file_name && (f->flags & FIO_FILE_UNLINK) &&
+		    f->filetype == FIO_TYPE_FILE) {
 			unlink(f->file_name);
 			free(f->file_name);
 			f->file_name = NULL;
 		}
 
-		close_file(td, f);
+		td_io_close_file(td, f);
 
-		if (f->mmap) {
-			munmap(f->mmap, f->file_size);
-			f->mmap = NULL;
-		}
+		if (f->file_map)
+			free(f->file_map);
 	}
 
 	td->filename = NULL;
-	free(td->files);
 	td->files = NULL;
 	td->nr_files = 0;
 }
+
+static void get_file_type(struct fio_file *f)
+{
+	struct stat sb;
+
+	f->filetype = FIO_TYPE_FILE;
+
+	if (!lstat(f->file_name, &sb)) {
+		f->flags |= FIO_FILE_EXISTS;
+
+		if (S_ISBLK(sb.st_mode))
+			f->filetype = FIO_TYPE_BD;
+		else if (S_ISCHR(sb.st_mode))
+			f->filetype = FIO_TYPE_CHAR;
+		else {
+			/*
+			 * might as well do this here, and save a stat later on
+			 */
+			f->real_file_size = sb.st_size;
+			f->file_size = f->real_file_size;
+		}
+	}
+}
+
+void add_file(struct thread_data *td, const char *fname)
+{
+	int cur_files = td->files_index;
+	struct fio_file *f;
+
+	td->files = realloc(td->files, (cur_files + 1) * sizeof(*f));
+
+	f = &td->files[cur_files];
+	memset(f, 0, sizeof(*f));
+	f->fd = -1;
+	f->file_name = strdup(fname);
+
+	get_file_type(f);
+
+	td->files_index++;
+	if (f->filetype == FIO_TYPE_FILE)
+		td->nr_normal_files++;
+}
+
+void get_file(struct fio_file *f)
+{
+	f->references++;
+}
+
+void put_file(struct thread_data *td, struct fio_file *f)
+{
+	if (!(f->flags & FIO_FILE_OPEN))
+		return;
+
+	assert(f->references);
+	if (--f->references)
+		return;
+
+	if (should_fsync(td) && td->fsync_on_close)
+		fsync(f->fd);
+
+	if (td->io_ops->close_file)
+		td->io_ops->close_file(td, f);
+	td->nr_open_files--;
+	f->flags &= ~FIO_FILE_OPEN;
+}
+
+static int recurse_dir(struct thread_data *td, const char *dirname)
+{
+	struct dirent *dir;
+	int ret = 0;
+	DIR *D;
+
+	D = opendir(dirname);
+	if (!D) {
+		td_verror(td, errno, "opendir");
+		return 1;
+	}
+
+	while ((dir = readdir(D)) != NULL) {
+		char full_path[PATH_MAX];
+		struct stat sb;
+
+		if (!strcmp(dir->d_name, ".") || !strcmp(dir->d_name, ".."))
+			continue;
+
+		sprintf(full_path, "%s/%s", dirname, dir->d_name);
+
+		if (lstat(full_path, &sb) == -1) {
+			if (errno != ENOENT) {
+				td_verror(td, errno, "stat");
+				return 1;
+			}
+		}
+
+		if (S_ISREG(sb.st_mode)) {
+			add_file(td, full_path);
+			td->nr_files++;
+			continue;
+		}
+
+		if ((ret = recurse_dir(td, full_path)) != 0)
+			break;
+	}
+
+	closedir(D);
+	return ret;
+}
+
+int add_dir_files(struct thread_data *td, const char *path)
+{
+	return recurse_dir(td, path);
+}
