@@ -17,6 +17,7 @@
 #include "crc32.h"
 #include "arch.h"
 #include "os.h"
+#include "mutex.h"
 
 #ifdef FIO_HAVE_SYSLET
 #include "syslet.h"
@@ -211,11 +212,17 @@ enum fio_filetype {
 
 enum fio_ioengine_flags {
 	FIO_SYNCIO	= 1 << 0,	/* io engine has synchronous ->queue */
-	FIO_CPUIO	= 1 << 1,	/* cpu burner, doesn't do real io */
-	FIO_MMAPIO	= 1 << 2,	/* uses memory mapped io */
-	FIO_RAWIO	= 1 << 3,	/* some sort of direct/raw io */
-	FIO_DISKLESSIO	= 1 << 4,	/* no disk involved */
-	FIO_SELFOPEN	= 1 << 5,       /* opens its own devices */
+	FIO_RAWIO	= 1 << 1,	/* some sort of direct/raw io */
+	FIO_DISKLESSIO	= 1 << 2,	/* no disk involved */
+	FIO_NOEXTEND	= 1 << 3,	/* engine can't extend file */
+	FIO_NODISKUTIL  = 1 << 4,       /* diskutil can't handle filename */
+};
+
+enum fio_file_flags {
+	FIO_FILE_OPEN		= 1 << 0,	/* file is open */
+	FIO_FILE_UNLINK		= 1 << 1,	/* unlink on close */
+	FIO_FILE_CLOSING	= 1 << 2,	/* file being closed */
+	FIO_FILE_EXISTS		= 1 << 3,	/* no need to create */
 };
 
 /*
@@ -223,6 +230,8 @@ enum fio_ioengine_flags {
  * this structure holds state information for a single file.
  */
 struct fio_file {
+	enum fio_filetype filetype;
+
 	/*
 	 * A file may not be a file descriptor, let the io engine decide
 	 */
@@ -245,10 +254,25 @@ struct fio_file {
 	unsigned int num_maps;
 	unsigned int last_free_lookup;
 
-	unsigned int unlink;
+	int references;
+	enum fio_file_flags flags;
 };
 
+/*
+ * How many depth levels to log
+ */
+#define FIO_IO_U_MAP_NR	8
+#define FIO_IO_U_LAT_NR 12
+
 struct thread_stat {
+	char *name;
+	char *verror;
+	int error;
+	int groupid;
+	pid_t pid;
+	char *description;
+	int members;
+
 	struct io_log *slat_log;
 	struct io_log *clat_log;
 	struct io_log *bw_log;
@@ -271,13 +295,18 @@ struct thread_stat {
 	unsigned long usr_time;
 	unsigned long sys_time;
 	unsigned long ctx;
-};
 
-/*
- * How many depth levels to log
- */
-#define FIO_IO_U_MAP_NR	8
-#define FIO_IO_U_LAT_NR 12
+	/*
+	 * IO depth and latency stats
+	 */
+	unsigned int io_u_map[FIO_IO_U_MAP_NR];
+	unsigned int io_u_lat[FIO_IO_U_LAT_NR];
+	unsigned long total_io_u[2];
+
+	unsigned long long io_bytes[2];
+	unsigned long runtime[2];
+	unsigned long total_run_time;
+};
 
 /*
  * This describes a single thread/process executing a fio job.
@@ -288,16 +317,17 @@ struct thread_data {
 	char *name;
 	char *directory;
 	char *filename;
+	char *ioengine;
 	char verror[128];
 	pthread_t thread;
 	int thread_number;
 	int groupid;
 	struct thread_stat ts;
-	enum fio_filetype filetype;
 	struct fio_file *files;
+	unsigned int files_index;
 	unsigned int nr_files;
 	unsigned int nr_open_files;
-	unsigned int nr_uniq_files;
+	unsigned int nr_normal_files;
 	union {
 		unsigned int next_file;
 		os_random_state_t next_file_state;
@@ -328,6 +358,7 @@ struct thread_data {
 	unsigned int write_bw_log;
 	unsigned int norandommap;
 	unsigned int bs_unaligned;
+	unsigned int fsync_on_close;
 
 	unsigned int bs[2];
 	unsigned int min_bs[2];
@@ -361,6 +392,9 @@ struct thread_data {
 	unsigned int rwmixwrite;
 	unsigned int nice;
 	unsigned int file_service_type;
+	unsigned int group_reporting;
+	unsigned int open_files;
+	char *opendir;
 
 	char *read_iolog_file;
 	char *write_iolog_file;
@@ -385,9 +419,6 @@ struct thread_data {
 	 * Current IO depth and list of free and busy io_u's.
 	 */
 	unsigned int cur_depth;
-	unsigned int io_u_map[FIO_IO_U_MAP_NR];
-	unsigned int io_u_lat[FIO_IO_U_LAT_NR];
-	unsigned long total_io_u;
 	struct list_head io_u_freelist;
 	struct list_head io_u_busylist;
 	struct list_head io_u_requeues;
@@ -404,7 +435,6 @@ struct thread_data {
 	unsigned long rate_bytes;
 	struct timeval lastrate;
 
-	unsigned long runtime[2];		/* msec */
 	unsigned long long io_size;
 	unsigned long long total_file_size;
 	unsigned long long start_offset;
@@ -415,7 +445,7 @@ struct thread_data {
 	unsigned long long io_bytes[2];
 	unsigned long long this_io_bytes[2];
 	unsigned long long zone_bytes;
-	volatile int mutex;
+	struct fio_sem *mutex;
 
 	/*
 	 * State for random io, a bitmap of blocks done vs not done
@@ -430,7 +460,6 @@ struct thread_data {
 
 	struct timeval start;	/* start of this loop */
 	struct timeval epoch;	/* time job was started */
-	struct timeval end_time;/* time job ended */
 
 	/*
 	 * read/write mixed workload state
@@ -456,6 +485,13 @@ struct thread_data {
 	 */
 	struct timeval timeout_end;
 	struct itimerval timer;
+
+	/*
+	 * for fileservice, how often to switch to a new file
+	 */
+	unsigned int file_service_nr;
+	unsigned int file_service_left;
+	struct fio_file *file_service_file;
 };
 
 /*
@@ -490,6 +526,7 @@ enum {
 
 extern int exitall_on_terminate;
 extern int thread_number;
+extern int nr_process, nr_thread;
 extern int shm_id;
 extern int groupid;
 extern int terse_output;
@@ -497,6 +534,7 @@ extern FILE *f_out;
 extern FILE *f_err;
 extern int temp_stall_ts;
 extern unsigned long long mlock_size;
+extern unsigned long page_mask, page_size;
 
 extern struct thread_data *threads;
 
@@ -619,9 +657,13 @@ extern int __must_check init_random_state(struct thread_data *);
 extern void close_files(struct thread_data *);
 extern int __must_check setup_files(struct thread_data *);
 extern int __must_check open_files(struct thread_data *);
-extern int open_file(struct thread_data *, struct fio_file *, int, int);
-extern void close_file(struct thread_data *, struct fio_file *);
 extern int __must_check file_invalidate_cache(struct thread_data *, struct fio_file *);
+extern int __must_check generic_open_file(struct thread_data *, struct fio_file *);
+extern void generic_close_file(struct thread_data *, struct fio_file *);
+extern void add_file(struct thread_data *, const char *);
+extern void get_file(struct fio_file *);
+extern void put_file(struct thread_data *, struct fio_file *);
+extern int add_dir_files(struct thread_data *, const char *);
 
 /*
  * ETA/status stuff
@@ -672,8 +714,10 @@ extern void requeue_io_u(struct thread_data *, struct io_u **);
 extern long __must_check io_u_sync_complete(struct thread_data *, struct io_u *);
 extern long __must_check io_u_queued_complete(struct thread_data *, int);
 extern void io_u_queued(struct thread_data *, struct io_u *);
+extern void io_u_log_error(struct thread_data *, struct io_u *);
 extern void io_u_init_timeout(void);
 extern void io_u_set_timeout(struct thread_data *);
+extern void io_u_mark_depth(struct thread_data *, struct io_u *);
 
 /*
  * io engine entry points
@@ -684,30 +728,8 @@ extern int __must_check td_io_queue(struct thread_data *, struct io_u *);
 extern int __must_check td_io_sync(struct thread_data *, struct fio_file *);
 extern int __must_check td_io_getevents(struct thread_data *, int, int, struct timespec *);
 extern int __must_check td_io_commit(struct thread_data *);
-
-/*
- * This is a pretty crappy semaphore implementation, but with the use that fio
- * has (just signalling start/go conditions), it doesn't have to be better.
- * Naturally this would not work for any type of contended semaphore or
- * for real locking.
- */
-static inline void fio_sem_init(volatile int *sem, int val)
-{
-	*sem = val;
-}
-
-static inline void fio_sem_down(volatile int *sem)
-{
-	while (*sem == 0)
-		usleep(10000);
-
-	(*sem)--;
-}
-
-static inline void fio_sem_up(volatile int *sem)
-{
-	(*sem)++;
-}
+extern int __must_check td_io_open_file(struct thread_data *, struct fio_file *);
+extern void td_io_close_file(struct thread_data *, struct fio_file *);
 
 /*
  * If logging output to a file, stderr should go to both stderr and f_err
@@ -735,12 +757,13 @@ struct ioengine_ops {
 	struct io_u *(*event)(struct thread_data *, int);
 	int (*cancel)(struct thread_data *, struct io_u *);
 	void (*cleanup)(struct thread_data *);
+	int (*open_file)(struct thread_data *, struct fio_file *);
+	void (*close_file)(struct thread_data *, struct fio_file *);
 	void *data;
 	void *dlhandle;
-	unsigned long priv;
 };
 
-#define FIO_IOOPS_VERSION	5
+#define FIO_IOOPS_VERSION	6
 
 extern struct ioengine_ops *load_ioengine(struct thread_data *, const char *);
 extern void register_ioengine(struct ioengine_ops *);
@@ -757,7 +780,7 @@ extern void close_ioengine(struct thread_data *);
 #define for_each_td(td, i)	\
 	for ((i) = 0, (td) = &threads[0]; (i) < (int) thread_number; (i)++, (td)++)
 #define for_each_file(td, f, i)	\
-	for ((i) = 0, (f) = &(td)->files[0]; (i) < (int) (td)->nr_files; (i)++, (f)++)
+	for ((i) = 0, (f) = &(td)->files[0]; (i) < (td)->nr_files; (i)++, (f)++)
 
 #define fio_assert(td, cond)	do {	\
 	if (!(cond)) {			\
