@@ -28,12 +28,6 @@ static int check_engine_ops(struct ioengine_ops *ops)
 		return 1;
 	}
 
-	/*
-	 * cpu thread doesn't need to provide anything
-	 */
-	if (ops->flags & FIO_CPUIO)
-		return 0;
-
 	if (!ops->queue) {
 		log_err("%s: no queue handler\n", ops->name);
 		return 1;
@@ -158,10 +152,7 @@ void close_ioengine(struct thread_data *td)
 	if (td->io_ops->dlhandle)
 		dlclose(td->io_ops->dlhandle);
 
-#if 0
-	/* we can't do this for threads, so just leak it, it's exiting */
 	free(td->io_ops);
-#endif
 	td->io_ops = NULL;
 }
 
@@ -195,6 +186,11 @@ int td_io_queue(struct thread_data *td, struct io_u *io_u)
 	assert((io_u->flags & IO_U_F_FLIGHT) == 0);
 	io_u->flags |= IO_U_F_FLIGHT;
 
+	assert(io_u->file->flags & FIO_FILE_OPEN);
+
+	io_u->error = 0;
+	io_u->resid = 0;
+
 	if (td->io_ops->flags & FIO_SYNCIO) {
 		fio_gettime(&io_u->issue_time, NULL);
 
@@ -208,8 +204,13 @@ int td_io_queue(struct thread_data *td, struct io_u *io_u)
 	if (io_u->ddir != DDIR_SYNC)
 		td->io_issues[io_u->ddir]++;
 
+	io_u_mark_depth(td, io_u);
+
 	ret = td->io_ops->queue(td, io_u);
 
+	if (ret == FIO_Q_QUEUED || ret == FIO_Q_COMPLETED)
+		get_file(io_u->file);
+
 	if (ret == FIO_Q_QUEUED) {
 		int r;
 
@@ -254,3 +255,33 @@ int td_io_commit(struct thread_data *td)
 
 	return 0;
 }
+
+int td_io_open_file(struct thread_data *td, struct fio_file *f)
+{
+	if (td->io_ops->open_file(td, f))
+		return 1;
+
+	f->last_free_lookup = 0;
+	f->last_completed_pos = 0;
+	f->last_pos = 0;
+	f->flags |= FIO_FILE_OPEN;
+	f->flags &= ~FIO_FILE_CLOSING;
+
+	if (f->file_map)
+		memset(f->file_map, 0, f->num_maps * sizeof(long));
+
+	td->nr_open_files++;
+	get_file(f);
+	return 0;
+}
+
+void td_io_close_file(struct thread_data *td, struct fio_file *f)
+{
+	/*
+	 * mark as closing, do real close when last io on it has completed
+	 */
+	f->flags |= FIO_FILE_CLOSING;
+
+	put_file(td, f);
+}
+
